__version__ = '1.1.2'

import tbats.abstract as abstract
import tbats.bats as bats
import tbats.tbats as tbats
from .bats import BATS
from .tbats import TBATS

__all__ = ['BATS', 'TBATS',
           'bats', 'tbats',
           'abstract']



from ..abstract import Case as AbstractCase


class Case(AbstractCase):

    def fit_initial_model(self, y):
        """Fits model with no ARMA and returns it

        TBATS, unlike BATS, does not check for non-seasonal model,
        non-seasonal models for TBATS are checked when seasonal harmonics are being determined

        Parameters
        ----------
        y: array-like
            Time series

        Returns
        -------
        Model
            Fitted model with no ARMA
        """
        return self.fit_case(y, self.components.without_arma())



import numpy as np

from . import MatrixBuilderInterface
from . import ArrayHelper


class MatrixBuilder(MatrixBuilderInterface):

    def __init__(self, model_params):
        self.params = model_params

    def calculate_D_matrix(self):
        F = self.make_F_matrix()
        g = self.make_g_vector()
        w = self.make_w_vector()
        D = F - np.outer(g, w)
        return D

    def make_seasonal_components_for_w(self):
        raise NotImplementedError()

    def make_gamma_vector(self):
        raise NotImplementedError()

    def make_A_matrix(self):
        raise NotImplementedError()

    def seasonal_components_amount(self):
        return self.params.seasonal_components_amount()

    def make_w_vector(self):
        ar_coefs = self.params.ar_coefs
        ma_coefs = self.params.ma_coefs

        w = np.concatenate([
            [1.0],
            [self.params.phi],
            self.make_seasonal_components_for_w(),
            ar_coefs,
            ma_coefs
        ])
        w = w[w != np.array(None)]
        return np.asarray(w, dtype=float)

    def make_g_vector(self):
        g = np.concatenate([
            [self.params.alpha],
            [self.params.beta],
            self.make_gamma_vector(),
            ArrayHelper.make_one_and_zeroes_vector(self.params.components.p),
            ArrayHelper.make_one_and_zeroes_vector(self.params.components.q),
        ])
        g = g[g != np.array(None)]
        return np.asarray(g, dtype=float)

    def make_F_matrix(self):
        ar_coefs = self.params.ar_coefs
        ma_coefs = self.params.ma_coefs
        phi = ArrayHelper.to_array(self.params.phi)
        seasonal_periods = self.params.components.seasonal_periods

        tao = self.seasonal_components_amount()
        p = len(ar_coefs)
        q = len(ma_coefs)

        has_beta = self.params.components.use_trend
        has_phi = self.params.components.use_trend  # Even if use_damped_trend=False phi is used and fixed to 1.0
        has_seasonal = (tao > 0)

        F = []

        row = [np.ones((1, 1))]
        if has_phi:
            row.append(phi)
        if has_seasonal:
            row.append(np.zeros((1, tao)))
        if p > 0:
            row.append(self.params.alpha * ar_coefs)
        if q > 0:
            row.append(self.params.alpha * ma_coefs)
        F.append(row)

        if has_beta:
            row = [np.zeros((1, 1))]
            if has_phi:
                row.append(phi)
            if has_seasonal:
                row.append(np.zeros((1, tao)))
            if p > 0:
                row.append(self.params.beta * ar_coefs)
            if q > 0:
                row.append(self.params.beta * ma_coefs)
            F.append(row)

        if has_seasonal:
            row = [np.zeros((tao, 1))]
            if has_phi:
                row.append(np.zeros((tao, 1)))
            row.append(self.make_A_matrix())  # A matrix
            gamma_vectors = self.make_gamma_vector()
            if p > 0:
                row.append(np.outer(gamma_vectors, ar_coefs))  # B matrix
            if q > 0:
                row.append(np.outer(gamma_vectors, ma_coefs))  # C matrix
            F.append(row)

        if p > 0:
            row = [np.zeros((p, 1))]
            if has_phi:
                row.append(np.zeros((p, 1)))
            if has_seasonal:
                row.append(np.zeros((p, tao)))
            ar_matrix = np.block([[ar_coefs], [np.eye(p - 1 if p > 0 else 0, p)]])
            row.append(ar_matrix)
            if q > 0:
                ar_matrix_ma_part = np.block([[ma_coefs], [np.zeros((p - 1 if p > 0 else 0, q))]])
                row.append(ar_matrix_ma_part)
            F.append(row)

        if q > 0:
            row = [np.zeros((q, 1))]
            if has_phi:
                row.append(np.zeros((q, 1)))
            if has_seasonal:
                row.append(np.zeros((q, tao)))
            if p > 0:
                ma_matrix_ar_part = np.zeros((q, p))
                row.append(ma_matrix_ar_part)
            ma_matrix = np.eye(q, q, -1)
            row.append(ma_matrix)
            F.append(row)

        return np.block(F)



from .ArrayHelper import ArrayHelper
from .MatrixBuilderInterface import MatrixBuilderInterface
from .MatrixBuilderCache import MatrixBuilderCache
from .MatrixBuilder import MatrixBuilder
from .Model import Model
from .ModelParams import ModelParams
from .Components import Components
from .ComponentMatrix import ComponentMatrix
from .ParamsOptimizer import ParamsOptimizer
from .SeedFinder import SeedFinder
from .Case import Case
from .Estimator import Estimator
from .ContextInterface import ContextInterface
from .Context import Context

__all__ = ['Model', 'MatrixBuilderInterface', 'MatrixBuilderCache', 'MatrixBuilder', 'ModelParams', 'ArrayHelper',
           'Components', 'ParamsOptimizer', 'SeedFinder', 'Case', 'ContextInterface', 'Estimator', 'ComponentMatrix',
           'Context']



import numpy as np

from ..abstract import ArrayHelper, ModelParams as AbstractModelParams


class ModelParams(AbstractModelParams):
    """Holds all parameters needed to calculate BATS model and forecasts

        For other parameters documentation see parent class documentation.

        Parameters
        ----------
        gamma_params: array-like of floats
            Gamma smoothing parameters for seasonal effects.
            In BATS there is one gamma parameter for each season.
        """

    def __init__(self, components, alpha, beta=None, phi=None, box_cox_lambda=None,
                 gamma_params=None,
                 ar_coefs=None, ma_coefs=None, x0=None):
        """See parent class documentation for details"""
        super().__init__(components=components, alpha=alpha, beta=beta, phi=phi, box_cox_lambda=box_cox_lambda,
                         gamma_params=gamma_params, ar_coefs=ar_coefs, ma_coefs=ma_coefs, x0=x0)

    @classmethod
    def with_default_starting_params(cls, y, components):
        """Factory method for starting model parameters.

        Parameters
        ----------
        y: array-like of floats
            Time series
        components: Components
            Components used in the model

        Returns
        -------
        ModelParams
            Default starting params.
        """
        alpha = 0.09
        if components.seasonal_periods.sum() > 16:
            alpha = 1e-6

        beta = None
        phi = None
        if components.use_trend:
            beta = 0.05
            if components.seasonal_periods.sum() > 16:
                beta = 5e-7
            phi = 1
            if components.use_damped_trend:
                phi = 0.999

        gamma_params = None
        if len(components.seasonal_periods) > 0:
            gamma_params = [0.001] * components.gamma_params_amount()

        box_cox_lambda = cls.find_initial_box_cox_lambda(y, components)

        # note that ARMA and x0 will be initialized to 0's in constructor
        return cls(components=components,
                   alpha=alpha, beta=beta, phi=phi,
                   gamma_params=gamma_params,
                   box_cox_lambda=box_cox_lambda)

    def seasonal_components_amount(self):
        """BATS model requires this many seasonal seed state values"""
        return int(self.components.seasonal_periods.sum())

    def _normalize_gamma_params(self, gamma_params=None):
        """Ensures gamma params is an array. If necessary initializes Gamma params to zeroes"""
        gamma_params = ArrayHelper.to_array(gamma_params)
        if len(gamma_params) != len(self.components.seasonal_periods):
            gamma_params = np.asarray([0.0] * len(self.components.seasonal_periods), float)
        return gamma_params



import numpy as np

from ..abstract import MatrixBuilder as AbstractMatrixBuilder
from . import ModelParams


class MatrixBuilder(AbstractMatrixBuilder):

    def __init(self, model_params):
        """

        :param ModelParams model_params:
        :return:
        """
        super().__init__(model_params)

    def make_seasonal_components_for_w(self):
        seasonal_components = []
        for harmonic in self.params.components.seasonal_harmonics:
            if harmonic > 0:
                seasonal_components.append([1.0] * harmonic)
                seasonal_components.append([0.0] * harmonic)
        if len(seasonal_components) > 0:
            seasonal_components = np.concatenate(seasonal_components)
        return seasonal_components

    def make_gamma_vector(self):
        gamma_1 = self.params.gamma_1()
        gamma_2 = self.params.gamma_2()
        seasonal_harmonics = self.params.components.seasonal_harmonics
        gamma_vectors = []
        for i in range(0, len(seasonal_harmonics)):  # assertion: seasonal_harmonics length equals amount of periods
            k = seasonal_harmonics[i]
            gamma_vectors.append([gamma_1[i]] * k)
            gamma_vectors.append([gamma_2[i]] * k)
        if len(gamma_vectors) > 0:
            gamma_vectors = np.concatenate(gamma_vectors)
        return np.asarray(gamma_vectors)

    def make_A_matrix(self):
        seasonal_periods = self.params.components.seasonal_periods
        seasonal_harmonics = self.params.components.seasonal_harmonics
        if len(seasonal_periods) == 0:
            return np.zeros((0, 0))
        A_array = []
        for period in range(0, len(seasonal_periods)):
            harmonics_amount = seasonal_harmonics[period]
            if harmonics_amount == 0:
                continue
            period_length = seasonal_periods[period]
            period_lambda = 2 * np.pi * np.asarray(range(1, harmonics_amount + 1)) / period_length
            cos_coefs = np.diag(np.cos(period_lambda))
            sin_coefs = np.diag(np.sin(period_lambda))
            Ai = np.block([
                [cos_coefs, sin_coefs],
                [-sin_coefs, cos_coefs],
            ])
            A_array.append(Ai)
        tao = int(2 * np.sum(seasonal_harmonics))
        dsum = np.zeros((tao, tao))
        offsetRow = 0
        offsetCol = 0
        for Ai in A_array:
            dsum[offsetRow:offsetRow + Ai.shape[0], offsetCol:offsetCol + Ai.shape[1]] = Ai
            offsetRow += Ai.shape[0]
            offsetCol += Ai.shape[1]
        return dsum



import copy
import numpy as np


class Components(object):
    """Contains all the information necessary to determine the amount and the structure of parameters

    Parameters
    ----------
    use_box_cox: bool
        If box cox will be used
    box_cox_bounds: tuple, shape=(2,)
        Minimal and maximal allowed values for Box-Cox lambda transformation.
    use_trend: bool
        If trend will be used
    use_damped_trend: bool
        If trend damping will apply. Applicable only when use_trend=True
    seasonal_periods: array-like
        Seasonal period lengths (amount of observations for one period of each seasonality)
    use_arma_errors:
        If ARMA(p,q) residuals modelling shall be tried.
    p: int
        Amount of AR(p) terms for ARMA(p,q) residuals modelling.
        If use_arma_errors=False then p=0
    q: int
        Amount of MA(p) terms for ARMA(p,q) residuals modelling.
        If use_arma_errors=False then q=0
    """

    def __init__(self, use_box_cox=False, use_trend=False, use_damped_trend=False,
                 seasonal_periods=None, use_arma_errors=False, p=0, q=0,
                 box_cox_bounds=(0, 1)):
        """Contains all the information necessary to determine the amount and the structure of parameters

        Does not contain model parameter values

        Parameters
        ----------
        use_box_cox: bool, optional (default=False)
            If box cox will be used
        use_trend: bool, optional (default=False)
            If trend will be used
        use_damped_trend: bool, optional (default=False)
            If trend damping will apply. Applicable only when use_trend=True
        seasonal_periods: array-like or None, optional (default=None)
            Seasonal period lengths (amount of observations for one period of each seasonality)
            When None, there there will be no seasonal periods
        use_arma_errors: bool, optional (default=False)
            If ARMA(p,q) residuals modelling shall be tried.
        p: int, optional (default=0)
            Amount of AR(p) terms for ARMA(p,q) residuals modelling.
            If use_arma_errors=False then p=0
        q: int, optional (default=0)
            Amount of MA(p) terms for ARMA(p,q) residuals modelling.
            If use_arma_errors=False then q=0
        box_cox_bounds: tuple, shape=(2,), optional (default=(0, 1))
            Minimal and maximal allowed values for Box-Cox lambda transformation.
        """
        self.use_box_cox = use_box_cox
        self.box_cox_bounds = box_cox_bounds
        self.use_trend = use_trend
        self.use_damped_trend = False
        if self.use_trend:  # only when trend is used
            self.use_damped_trend = use_damped_trend
        self.seasonal_periods = self._normalize_seasons(seasonal_periods)
        self.__use_arma(do_use=use_arma_errors, p=p, q=q)

    @classmethod
    def create_constant_components(cls):
        # abstract method
        raise NotImplementedError()

    def _normalize_seasons(self, seasonal_periods):
        # abstract method
        raise NotImplementedError()

    def seasonal_components_amount(self):
        # abstract method
        raise NotImplementedError()

    def gamma_params_amount(self):
        # abstract method
        raise NotImplementedError()

    def _seasonal_summary(self):
        # abstract method
        raise NotImplementedError()

    def arma_length(self):
        """Returns amount of ARMA(p,q) parameters"""
        return self.p + self.q

    def summary(self):
        """Returns components summary

        Returns
        -------
        str
            summary of components
        """
        s = "Use Box-Cox: %r\n" % self.use_box_cox
        s += "Use trend: %r\n" % self.use_trend
        s += "Use damped trend: %r\n" % self.use_damped_trend
        s += "Seasonal periods: %s\n" % self.seasonal_periods
        s += self._seasonal_summary()
        s += "ARMA errors (p, q): (%d, %d)\n" % (self.p, self.q)
        return s

    def with_seasonal_periods(self, seasonal_periods):
        """Creates copy of itself but with new seasonal periods

        Parameters
        ----------
        seasonal_periods: array-like
            New season lengths

        Returns
        -------
        Components
            copy of components with new seasonal periods
        """
        components = copy.deepcopy(self)
        components.seasonal_periods = self._normalize_seasons(seasonal_periods)
        return components

    def without_seasonal_periods(self):
        """Creates copy itself without seasonality

        Returns
        -------
        Components
            copy of components without seasonal periods
        """
        components = copy.deepcopy(self)
        components.seasonal_periods = np.asarray([])
        return components

    def without_arma(self):
        """Creates copy itself without ARMA components

        Returns
        -------
        Components
            copy of components without ARMA
        """
        components = copy.deepcopy(self)
        return components.__use_arma(False)

    def with_arma(self, p=0, q=0):
        """Creates copy itself with provided ARMA components

        Returns
        -------
        Components
            copy of components with provided ARMA(p,q) degrees
        """
        components = copy.deepcopy(self)
        return components.__use_arma(do_use=True, p=p, q=q)

    def is_box_cox_in_bounds(self, box_cox_lambda):
        """Tells if provided Box-Cox lambda is within allowed interval

        When box cox is not used, returns True regardless of lambda value.

        Returns
        -------
        bool
            True when lambda is in bounds
        """
        if self.use_box_cox:
            return self.box_cox_bounds[0] <= box_cox_lambda <= self.box_cox_bounds[1]
        return True

    def __use_arma(self, do_use=True, p=0, q=0):
        """Modifies object state so that ARMA parameters according to provided"""
        self.use_arma_errors = do_use
        self.p = 0
        self.q = 0
        if do_use:  # only when ARMA errors are used
            self.p = p
            self.q = q
        return self



class MatrixBuilderInterface(object):

    def make_w_vector(self):
        raise NotImplementedError()

    def make_g_vector(self):
        raise NotImplementedError()

    def make_F_matrix(self):
        raise NotImplementedError()

    def calculate_D_matrix(self):
        raise NotImplementedError()



from .ModelParams import ModelParams
from .Components import Components
from .MatrixBuilder import MatrixBuilder
from .Model import Model
from .SeedFinder import SeedFinder
from .ParamsOptimizer import ParamsOptimizer
from .Case import Case
from .Context import Context
from .BATS import BATS

__all__ = ['BATS', 'ParamsOptimizer', 'Case', 'Model', 'Components', 'ModelParams', 'MatrixBuilder', 'Context',
           'SeedFinder']



import re

import setuptools
import distutils.cmd
import distutils.log
import subprocess
import sys

from setuptools.command.test import test as TestCommand

# Getting version:
with open("tbats/__init__.py") as init_file:
    version = re.search("__version__ = \'(.*?)\'", init_file.read()).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args).append('test/'))
        sys.exit(errno)


class RTestCommand(TestCommand):
    description = 'Run comparison to R forecast package implementation of BATS and TBATS (REQUIRES R)'

    def run_tests(self):
        """Run command."""
        command = ['pytest', 'test_R']
        # if self.pylint_rcfile:
        #     command.append('--rcfile=%s' % self.pylint_rcfile)
        # command.append(os.getcwd())
        self.announce(
            'Running command: %s' % str(command),
            level=distutils.log.INFO)
        subprocess.check_call(command)

setuptools.setup(
    name='tbats',
    version=version,
    packages=setuptools.find_packages(exclude=('test', 'test_R')),
    url='https://github.com/intive-DataScience/tbats',
    license='MIT License',
    author='Grzegorz Skorupa (intive)',
    author_email='grzegorz.skorupa@intive.com',
    description='BATS and TBATS for time series forecasting',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy', 'scipy', 'pmdarima', 'scikit-learn'],
    tests_require=['pytest'],
    extras_require={
        'dev': ['pip-tools', 'pytest', 'rpy2'],
    },
    cmdclass={
        'test': PyTest,
        'test_r': RTestCommand,
    },

)



from . import MatrixBuilderInterface


class MatrixBuilderCache(MatrixBuilderInterface):

    def __init__(self, matrix_builder):
        """
        Wraps matrix_builder so that all matrices are calculated once only

        :param MatrixBuilderInterface matrix_builder: The builder that provides matrices
        """
        self.builder = matrix_builder
        self.w = None
        self.g = None
        self.F = None
        self.D = None

    def make_w_vector(self):
        if self.w is None:
            self.w = self.builder.make_w_vector()
        return self.w

    def make_g_vector(self):
        if self.g is None:
            self.g = self.builder.make_g_vector()
        return self.g

    def make_F_matrix(self):
        if self.F is None:
            self.F = self.builder.make_F_matrix()
        return self.F

    def calculate_D_matrix(self):
        if self.D is None:
            self.D = self.builder.calculate_D_matrix()
        return self.D
